from typing import Any, Literal, TypedDict

DocumentValidationViolation = Literal[
    "INVALID_TYPE", "INVALID_LENGTH", "INVALID_VALUE", "INVALID_CHECK_DIGIT"
]


class _PerformValidationOkResult(TypedDict):
    ok: Literal[True]
    value: str


class PerformValidationErrorResult(TypedDict):
    ok: Literal[False]
    violation: DocumentValidationViolation
    message: str
    extra: dict[str, Any]


PerformValidationResult = PerformValidationErrorResult | _PerformValidationOkResult



from yabra.doc import types


class InvalidDocumentError(Exception):
    result: types.PerformValidationErrorResult

    def __init__(self, result: types.PerformValidationErrorResult) -> None:
        self.result = result
        super().__init__(result["message"])



import itertools
import re
from collections.abc import Iterable

from yabra.doc.base_doc import BaseDoc, Mod11CheckDigitsSpecification


class AlphaNumericCNPJSpecification(Mod11CheckDigitsSpecification):
    """Alpha-numeric CNPJ specification, validates both numeric and alpha-numeric documents"""

    MAX_LENGTH = 14
    NUM_OF_CHECK_DIGITS = 2
    VALID_DIGITS_RE = re.compile(r"[^A-Za-z0-9]")
    DEFAULT_FILTERED_VALUE_MASK = "{0}.***.{2}/****-{4}"

    @staticmethod
    def digit_value_for_check_digit(digit: str) -> int:
        # Ascii value of the digit - constant value 48
        return ord(digit.upper()) - 48

    @staticmethod
    def digit_weights_iterable(n_of_current_digits: int) -> Iterable[int]:
        # The weights starts from 2 and increase up to 9 (range), and start over (cycle).
        # The weight starts from right to left (reversed)
        return reversed(list(itertools.islice(itertools.cycle(range(2, 10)), n_of_current_digits)))


class NumericCNPJSpecification(AlphaNumericCNPJSpecification):
    """Accepts only the numeric form of the CNPJ document, the key difference is how the value is calculated
    for the check digits, before it's way simpler"""

    VALID_DIGITS_RE = re.compile(r"\D")

    @staticmethod
    def digit_value_for_check_digit(digit: str) -> int:
        return int(digit)


class CNPJ(BaseDoc):
    spec = AlphaNumericCNPJSpecification

    def masked_value(self) -> str:
        cnpj = self.validated_value()
        return f"{cnpj[:2]}.{cnpj[2:5]}.{cnpj[5:8]}/{cnpj[8:12]}-{cnpj[12:14]}"

    def filtered_value(self, mask: str | None = None) -> str:
        mask = mask or self.spec.DEFAULT_FILTERED_VALUE_MASK
        cnpj = self.validated_value()
        return mask.format(*[cnpj[:2], cnpj[2:5], cnpj[5:8], cnpj[8:12], cnpj[12:14]])



import re
from collections.abc import Iterable

from yabra.doc.base_doc import BaseDoc, Mod11CheckDigitsSpecification


class CPFSpecification(Mod11CheckDigitsSpecification):
    MAX_LENGTH = 11
    NUM_OF_CHECK_DIGITS = 2
    VALID_DIGITS_RE = re.compile(r"\D")
    DEFAULT_FILTERED_VALUE_MASK = "{0}.***.***-{3}"

    @staticmethod
    def digit_value_for_check_digit(digit: str) -> int:
        return int(digit)

    @staticmethod
    def digit_weights_iterable(n_of_current_digits: int) -> Iterable[int]:
        return range(n_of_current_digits + 1, 1, -1)


class CPF(BaseDoc):
    spec = CPFSpecification

    def masked_value(self) -> str:
        cpf = self.validated_value()
        return f"{cpf[:3]}.{cpf[3:6]}.{cpf[6:9]}-{cpf[9:11]}"

    def filtered_value(self, mask: str | None = None) -> str:
        mask = mask or self.spec.DEFAULT_FILTERED_VALUE_MASK
        cpf = self.validated_value()
        return mask.format(*[cpf[:3], cpf[3:6], cpf[6:9], cpf[9:11]])



from typing import get_args

from yabra.doc.generate_doc import (
    DocumentGenerator,
    SupportedDocTypes,
    SupportedMissingDigitsAlgorithm,
)

try:
    import click
except ImportError as e:
    raise ImportError("Cannot import `click`, make sure you install `yabra[cli]`") from e


@click.group()
def cli() -> None: ...


@cli.command(
    help="""Generates one or many valid documents""",
    no_args_is_help=True,
)
@click.argument(
    "doc-type",
    type=click.Choice(get_args(SupportedDocTypes)),
    nargs=1,
    required=True,
)
@click.option(
    "--prefix",
    "prefix",
    nargs=1,
    required=False,
    type=str,
    default="",
    help="""
The initial digits of the document. Can have up to doc_max_digits - doc_check_digits. For example,
cpf can have up to 9 chars and cnpj up to 12""",
)
@click.option(
    "--algorithm",
    "algorithm",
    nargs=1,
    required=False,
    type=click.Choice(get_args(SupportedMissingDigitsAlgorithm)),
    default="random",
    help="""
The algorithm that is going to be used to generate the missing digits, defaults to `random`
where missing digits is the difference between the maximum number digits for the given
document type - the prefix digits. We recommend that you use the `sequential` like algorithms
if generating a high number of documents in order to get the expected number of documents,
since the `random` algorithms aren't deterministic, you can get lesser documents than requested.""",
)
@click.option(
    "--number",
    "number",
    nargs=1,
    required=False,
    type=int,
    default=1,
    help="""
The number of documents to generate, defaults to 1. The maximum number of generatable documents depends on the
given prefix, it's also worth reading the note about the `algorithm` option when generating a large number of
documents (TLDR: Use `sequential` or `alpha_sequential`).""",
)
@click.option(
    "--mask",
    is_flag=True,
    default=False,
    help="""Controls if the output documents are masked or not, defaults to unmasked""",
)
def generate(
    doc_type: SupportedDocTypes,
    prefix: str,
    algorithm: SupportedMissingDigitsAlgorithm,
    number: int,
    mask: bool,
) -> None:
    try:
        generator = DocumentGenerator(
            doc_type=doc_type, algorithm=algorithm, prefix=prefix, number=number
        )
    except ValueError as e:
        click.echo(click.style("".join(e.args), "red"))
        raise click.exceptions.Exit(code=1)

    for doc in generator.iter_documents():
        click.echo(doc.masked_value() if mask else doc.validated_value())


if __name__ == "__main__":
    cli()



import itertools
import random
import string
from typing import Generator, Literal, Protocol, assert_never

from yabra.doc.base_doc import BaseDoc
from yabra.doc.cnpj import CNPJ
from yabra.doc.cpf import CPF


class MissingDigitsGenerator(Protocol):
    population: str

    def __init__(self) -> None: ...
    def get_digits(self, n: int) -> str: ...


class RandomAlphaNumericDigitsGenerator(MissingDigitsGenerator):
    def __init__(self) -> None:
        self.population = string.ascii_uppercase + string.digits

    def get_digits(self, n: int) -> str:
        return "".join(random.choices(self.population, k=n))


class RandomMissingDigitsGenerator(MissingDigitsGenerator):
    def __init__(self) -> None:
        self.population = string.digits

    def get_digits(self, n: int) -> str:
        return "".join(random.choices(self.population, k=n))


class SequentialMissingDigitsGenerator(MissingDigitsGenerator):
    def __init__(self) -> None:
        self.population = string.digits
        self.seq_range_iter = iter(range(999_999_999_999))

    def get_digits(self, n: int) -> str:
        digits = str(next(self.seq_range_iter))
        return digits.zfill(n)


class SequentialAlphaNumericMissingDigitsGenerator(MissingDigitsGenerator):
    def __init__(self) -> None:
        self.population = string.digits + string.ascii_uppercase
        self._iter = self._next_digit_iter()

    def get_digits(self, n: int) -> str:
        digits = next(self._iter)
        return digits.zfill(n)

    def _next_digit_iter(self) -> Generator[str, None, None]:
        length = 1
        while True:
            for combo in itertools.product(self.population, repeat=length):
                yield "".join(combo)
            length += 1


SupportedDocTypes = Literal["cpf", "cnpj"]


def get_doc_type_class(doc_type: SupportedDocTypes) -> type[BaseDoc]:
    if doc_type == "cpf":
        return CPF
    if doc_type == "cnpj":
        return CNPJ
    assert_never(doc_type)


SupportedMissingDigitsAlgorithm = Literal[
    "random", "alpha_random", "sequential", "alpha_sequential"
]


def get_missing_digits_algorithm_generator(
    algorithm: SupportedMissingDigitsAlgorithm,
) -> MissingDigitsGenerator:
    if algorithm == "random":
        return RandomMissingDigitsGenerator()
    if algorithm == "alpha_random":
        return RandomAlphaNumericDigitsGenerator()
    if algorithm == "sequential":
        return SequentialMissingDigitsGenerator()
    if algorithm == "alpha_sequential":
        return SequentialAlphaNumericMissingDigitsGenerator()
    assert_never(algorithm)


class DocumentGenerator:
    def __init__(
        self,
        doc_type: SupportedDocTypes,
        algorithm: SupportedMissingDigitsAlgorithm,
        prefix: str,
        number: int,
    ) -> None:
        self.doc_class = get_doc_type_class(doc_type)
        self.spec = self.doc_class.spec
        self.prefix_digits_max_length = self.spec.MAX_LENGTH - self.spec.NUM_OF_CHECK_DIGITS
        self.missing_digits_generator = self._get_valid_missing_digits_generator(algorithm)
        self.prefix = self._get_valid_prefix(prefix)
        self.number = self._get_valid_number(number)

    def _get_valid_missing_digits_generator(
        self, algorithm: SupportedMissingDigitsAlgorithm
    ) -> MissingDigitsGenerator:
        if algorithm == "random" or algorithm == "sequential":
            return get_missing_digits_algorithm_generator(algorithm)
        if algorithm == "alpha_random" or algorithm == "alpha_sequential":
            # Alpha numeric validation
            doc_allows_alpha_chars = self.spec.VALID_DIGITS_RE.sub("", "A")
            if not doc_allows_alpha_chars:
                raise ValueError("Incompatible algorithm for doc_type")

            return get_missing_digits_algorithm_generator(algorithm)
        assert_never(algorithm)

    def _get_valid_prefix(self, prefix: str) -> str:
        valid_prefix_digits = "".join(
            d for d in prefix if d in self.missing_digits_generator.population
        )
        if valid_prefix_digits != prefix:
            raise ValueError(
                f"Invalid prefix option, it contains invalid digits. Valid digits: {self.missing_digits_generator.population!r}",
                "You might want to verify if the provided prefix is compatible with the selected algorithm",
            )
        if len(valid_prefix_digits) > self.prefix_digits_max_length:
            raise ValueError(
                f"Invalid prefix option, it must not have length > {self.prefix_digits_max_length}"
            )
        return prefix

    def _get_valid_number(self, number: int) -> int:
        if number < 0:
            raise ValueError("Invalid number option, it must value must be greater than 0")

        doc_possibilities = len(self.missing_digits_generator.population) ** (
            self.prefix_digits_max_length - len(self.prefix)
        )
        if number > doc_possibilities:
            raise ValueError(
                f"Invalid number option, the required value is above the maximum number of possibilities ({doc_possibilities})"
            )
        return number

    def iter_documents(self) -> Generator[BaseDoc, None, None]:
        generated_docs = set[str]()
        iter_count = 0
        max_iter_count = self.number * 5
        # Useful for the random missing digits generator, since its not deterministic, we must define a maximum
        # number of iterations in order to not end up in a infinite loop

        while len(generated_docs) < self.number and iter_count < max_iter_count:
            iter_count += 1
            document = self._generate_one()
            if document.validated_value() in generated_docs:
                continue
            generated_docs.add(document.validated_value())
            yield document

    def _generate_one(self) -> BaseDoc:
        digits = self.prefix

        missing_digits = self.prefix_digits_max_length - len(digits)
        assert missing_digits >= 0

        if missing_digits:
            digits += self.missing_digits_generator.get_digits(n=missing_digits)
        return self.doc_class(self.spec.calculate_document_number(digits))



import abc
import re
from collections.abc import Iterable
from typing import Any, Never, assert_never

from yabra.doc import types
from yabra.doc.exc import InvalidDocumentError


class DocumentSpecification(abc.ABC):
    MAX_LENGTH: int
    NUM_OF_CHECK_DIGITS: int
    VALID_DIGITS_RE: re.Pattern[str]
    REPEATED_DIGITS_ALLOWED: bool = False
    DEFAULT_FILTERED_VALUE_MASK: str

    @staticmethod
    @abc.abstractmethod
    def digit_value_for_check_digit(digit: str) -> int: ...

    @staticmethod
    @abc.abstractmethod
    def digit_weights_iterable(n_of_current_digits: int) -> Iterable[int]: ...

    @classmethod
    @abc.abstractmethod
    def calculate_next_check_digit(cls, digits: str) -> str: ...

    @classmethod
    def calculate_document_number(cls, digits: str) -> str:
        digits_to_check = digits[: cls.MAX_LENGTH - cls.NUM_OF_CHECK_DIGITS]
        for _ in range(cls.NUM_OF_CHECK_DIGITS):
            digits_to_check += cls.calculate_next_check_digit(digits_to_check)

        assert len(digits_to_check) == cls.MAX_LENGTH
        return digits_to_check

    @staticmethod
    def parse_raw_value(value: Any) -> str | None: ...


class BaseDoc(abc.ABC):
    spec: type[DocumentSpecification]

    def __init__(self, value: Any) -> None:
        self.raw_value = value
        self._validated_value: str | None = None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.raw_value!r})"

    def __str__(self) -> str:
        if self._validated_value is not None:
            return self._validated_value
        return self.validated_value()

    def validated_value(self) -> str:
        if self._validated_value is not None:
            return self._validated_value

        result = self.perform_validation()
        if result["ok"] is True:
            self._validated_value = result["value"]
            return self._validated_value
        self.raise_invalid_document_error(result)
        assert_never("`raise_invalid_document_error` didn't raised an exception")

    def perform_validation(self) -> types.PerformValidationResult:
        value = self.spec.parse_raw_value(self.raw_value)
        if value is None:
            return {
                "ok": False,
                "violation": "INVALID_TYPE",
                "message": f"Documento inválido: Tipo {type(self.raw_value).__name__!r} não suportado",
                "extra": {"type": type(self.raw_value).__name__, "raw": repr(self.raw_value)},
            }
        digits = self.spec.VALID_DIGITS_RE.sub(repl="", string=value)
        if not digits:
            return {
                "ok": False,
                "violation": "INVALID_VALUE",
                "message": "Documento inválido: Nenhum dígito encontrado",
                "extra": {"value": value, "raw": repr(self.raw_value)},
            }
        if len(digits) != self.spec.MAX_LENGTH:
            return {
                "ok": False,
                "violation": "INVALID_LENGTH",
                "message": f"Documento inválido: Número de dígitos inválido, esperavam-se {self.spec.MAX_LENGTH} dígitos",
                "extra": {"expected_length": self.spec.MAX_LENGTH, "length": len(digits)},
            }

        if not self.spec.REPEATED_DIGITS_ALLOWED and len(set(digits)) == 1:
            return {
                "ok": False,
                "violation": "INVALID_VALUE",
                "message": "Documento inválido: Documento com números repetidos não é permitido",
                "extra": {"value": digits},
            }

        calculated_doc = self.spec.calculate_document_number(digits)
        if calculated_doc != digits:
            return {
                "ok": False,
                "violation": "INVALID_CHECK_DIGIT",
                "message": "Documento inválido: Os dígitos verificadores não coincidem",
                "extra": {"calculated": calculated_doc, "received": digits},
            }
        return {"ok": True, "value": digits.upper()}

    def raise_invalid_document_error(self, result: types.PerformValidationErrorResult) -> Never:
        raise InvalidDocumentError(result)

    @abc.abstractmethod
    def masked_value(self) -> str: ...

    @abc.abstractmethod
    def filtered_value(self, mask: str | None = None) -> str: ...


class Mod11CheckDigitsSpecification(DocumentSpecification):
    @classmethod
    def calculate_next_check_digit(cls, digits: str) -> str:
        weights = cls.digit_weights_iterable(len(digits))
        digits_sum = sum(
            cls.digit_value_for_check_digit(digit) * weight
            for digit, weight in zip(digits, weights, strict=True)
        )
        remainder = digits_sum % 11
        return "0" if remainder <= 1 else str(11 - remainder)

    @staticmethod
    def parse_raw_value(value: Any) -> str | None:
        if isinstance(value, str):
            return value
        if isinstance(value, bytes):
            return value.decode()
        if isinstance(value, int):
            return str(value)
        return None
